@@ -46,6 +46,7 @@ from smartcard.scard import \
     SCardGetErrorMessage as _SCardGetErrorMessage,\
     SCardGetStatusChange as _SCardGetStatusChange
 
+
 # Local application imports
 from common.intlist import to_hstr as _to_hstr
 from common.hstr import \
@@ -98,12 +99,12 @@ def establish_context(scope: Scope):
 
     hresult, hcontext = _SCardEstablishContext(dw_scope)
     if hresult != _SCARD_S_SUCCESS:
-        err = F'Failed to establish context: {_SCardGetErrorMessage(hresult)}'
+        err = F'Failed to establish context ({_SCardGetErrorMessage(hresult)})'
         logging.info(err)
-        return None, err
+        raise Exception(err)
 
     logging.info(F"Context established with scope={scope}")
-    return hcontext, None
+    return hcontext
 
 
 def list_readers(hcontext, readergroups=None):
@@ -114,12 +115,12 @@ def list_readers(hcontext, readergroups=None):
 
     hresult, readers = _SCardListReaders(hcontext, readergroups)
     if hresult != _SCARD_S_SUCCESS:
-        err = F'Failed to list readers: {_SCardGetErrorMessage(hresult)}'
+        err = F'Failed to list readers ({_SCardGetErrorMessage(hresult)})'
         logging.info(err)
-        return None, err
+        raise Exception(err)
 
     logging.info(F"Readers listed: {', '.join(readers)}")
-    return readers, None
+    return readers
 
 
 def connect(hcontext, reader, share_mode: ShareMode, preferred_protocols: Protocol):
@@ -131,13 +132,13 @@ def connect(hcontext, reader, share_mode: ShareMode, preferred_protocols: Protoc
     hresult, hcard, dw_active_protocol = _SCardConnect(
         hcontext, reader, dw_share_mode, dw_preferred_protocols)
     if hresult != _SCARD_S_SUCCESS:
-        err = F'Unable to connect: {_SCardGetErrorMessage(hresult)}'
+        err = F'Unable to connect ({_SCardGetErrorMessage(hresult)})'
         logging.info(err)
-        return None, err
+        raise Exception(err)
 
     protocol = Protocol(dw_active_protocol)
     logging.info(F"Connected with mode={share_mode}, protocol={protocol}")
-    return (hcard, protocol), None
+    return hcard, protocol
 
 
 def reconnect(hcard, share_mode: ShareMode, preferred_protocols: Protocol, initialization: Disposition):
@@ -151,14 +152,14 @@ def reconnect(hcard, share_mode: ShareMode, preferred_protocols: Protocol, initi
         hcard, dw_share_mode, dw_preferred_protocols, dw_initialization)
 
     if hresult != _SCARD_S_SUCCESS:
-        err = F'Unable to reconnect: {_SCardGetErrorMessage(hresult)}'
+        err = F'Unable to reconnect ({_SCardGetErrorMessage(hresult)})'
         logging.info(err)
-        return None, err
+        raise Exception(err)
 
     active_protocol = Protocol(dw_active_protocol)
     logging.info(
         F"Reconnected with mode={share_mode}, disposition={initialization}, protocol={active_protocol}")
-    return active_protocol, None
+    return active_protocol
 
 
 def status(hcard):
@@ -167,9 +168,9 @@ def status(hcard):
     hresult, reader, dw_state, dw_protocol, atr = _SCardStatus(hcard)
 
     if hresult != _SCARD_S_SUCCESS:
-        err = F'Unable to get current reader status: {_SCardGetErrorMessage(hresult)}'
+        err = F'Unable to get current reader status ({_SCardGetErrorMessage(hresult)})'
         logging.info(err)
-        return None, err
+        raise Exception(err)
 
     protocol = Protocol(dw_protocol)
     state = []
@@ -178,7 +179,7 @@ def status(hcard):
             state.append(scard_state)
     logging.info(
         F"Card status: reader='{reader}', state={state}, protocol={protocol}, ATR={_to_hstr(atr)}")
-    return (reader, state, protocol, _to_hstr(atr)), None
+    return reader, state, protocol, _to_hstr(atr)
 
 
 def transmit(hcard, protocol: Protocol, apdu: str):
@@ -191,21 +192,18 @@ def transmit(hcard, protocol: Protocol, apdu: str):
 
     hresult, response_bytes = _SCardTransmit(hcard, pio_send_pci, apdu_bytes)
     if hresult != _SCARD_S_SUCCESS:
-        err = F'Failed to transmit: {_SCardGetErrorMessage(hresult)}'
+        err = F'Failed to transmit ({_SCardGetErrorMessage(hresult)})'
         logging.info(err)
-        return None, err
+        raise Exception(err)
 
     response = _to_hstr(response_bytes)
     data = response[0:-4]
     SW1 = response[-4:-2]
     SW2 = response[-2:]
-    logging.info(F"Tx={apdu}")
-    if data == '':
-        logging.info(F"Rx={SW1}{SW2}")
-    else:
-        logging.info(F"Rx={data} {SW1}{SW2}")
+    logging.info(F"Tx->{apdu}")
+    logging.info(F"Rx<-{data}{SW1}{SW2}")
 
-    return (data, SW1, SW2), None
+    return data, SW1, SW2
 
 
 def send_apdu(hcard, protocol: Protocol, apdu: dict):
@@ -225,7 +223,7 @@ def send_apdu(hcard, protocol: Protocol, apdu: dict):
     Le = apdu.get('Le', None)
 
     if protocol not in Protocol:
-        return None, F'Unsupported protocol: {protocol}'
+        raise Exception(F'Unsupported protocol: {protocol}')
 
     if data is None:
         # Case 1
@@ -247,7 +245,7 @@ def send_apdu(hcard, protocol: Protocol, apdu: dict):
         logging.debug("Case 2E")
         if len(Le) == 6:
             if protocol == Protocol.T0:
-                return ['', '', ''], 'Unsupported case 2E with protocol T=0'
+                raise Exception('Unsupported case 2E with protocol T=0')
             if protocol == Protocol.T1:
                 return transmit(hcard, protocol, header + Le)
 
@@ -267,12 +265,12 @@ def send_apdu(hcard, protocol: Protocol, apdu: dict):
         if Lc < 65536:
             Lc = F"00{Lc:04X}"
             if protocol == Protocol.T0:
-                return ['', '', ''], 'Unsupported case 3E with protocol T=0'
+                raise Exception('Unsupported case 3E with protocol T=0')
             if protocol == Protocol.T1:
                 return transmit(hcard, protocol, header + Lc + data)
         # Data field too long
         else:
-            return ['', '', ''], F"Data field too long: {Lc:X}"
+            raise Exception(F"Data field too long: {Lc:X}")
 
     # Case 4S
     logging.debug("Case 4S")
@@ -285,11 +283,12 @@ def send_apdu(hcard, protocol: Protocol, apdu: dict):
     logging.debug("Case 4E")
     if Lc < 65536 and len(Le) == 6:
         if protocol == Protocol.T0:
-            return ['', '', ''], 'Unsupported case 4E with protocol T=0'
+            raise Exception('Unsupported case 4E with protocol T=0')
         if protocol == Protocol.T1:
             return transmit(hcard, protocol, header + F"00{Lc:04X}" + data + Le)
 
-    return ['', '', ''], F'Unsupported case with short and extended lengths: Lc = {Lc:X}, Le = {Le}'
+    raise Exception(
+        F'Unsupported case with short and extended lengths: Lc = {Lc:X}, Le = {Le}')
 
 
 def _send_apdu_T0_case_1(hcard, header: str):
@@ -306,23 +305,24 @@ def _send_apdu_T0_case_2s(hcard, header: str, Le: str):
         pass
     # Wrong Le field
     else:
-        return None, F'Wrong Le value: {Le}'
+        raise Exception(F'Wrong Le value: {Le}')
 
-    (data, SW1, SW2), _ = transmit(hcard, Protocol.T0, header + Le)
+    data, SW1, SW2 = transmit(hcard, Protocol.T0, header + Le)
     # Case 2S.1—Process completed: Ne accepted
     if SW1+SW2 == '9000':
-        return (data, SW1, SW2), None
+        return data, SW1, SW2
     # Case 2S.2—Process aborted: Ne definitively not accepted
     if SW1+SW2 == '6700':
-        return ('', SW1, SW2), 'Error—Process aborted: Ne definitively not accepted (6700)'
+        raise Exception(
+            'Error—Process aborted: Ne definitively not accepted (6700)')
     # Case 2S.3—Process aborted; Ne not accepted, Na indicated
     if SW1 == '6C':
         return transmit(hcard, Protocol.T0, header + SW2)
     # Case 2S.4—SW12 = '9XYZ', except for '9000'
     if SW1.startswith('9'):
-        return (data, SW1, SW2), None
+        return data, SW1, SW2
 
-    return [data, SW1, SW2], 'Unknown command case'
+    raise Exception(F'Unknown command case (data={data}, SW12={SW1}{SW2})')
 
 
 def _send_apdu_T0_case_3s(hcard, header: str, Lc: str, data: str):
@@ -333,7 +333,7 @@ def _send_apdu_T0_case_3s(hcard, header: str, Lc: str, data: str):
         pass
     # Data field too long
     else:
-        return None, F"Data field too long: {Lc}"
+        raise Exception(F"Data field too long: {Lc}")
 
     return transmit(hcard, Protocol.T0, header + Lc + data)
 
@@ -346,19 +346,20 @@ def _send_apdu_T0_case_4s(hcard, header: str, Lc: str, data: str, Le: str):
         pass
     # Data field too long
     else:
-        return None, F"Data field too long: {Lc}"
+        raise Exception(F"Data field too long: {Lc}")
 
     # Short Le field
     if len(Le) == 2:
         pass
     # Wrong Le field
     else:
-        return None, F'Wrong Le value: {Le}'
+        raise Exception(F'Wrong Le value: {Le}')
 
-    (data, SW1, SW2), _ = transmit(hcard, Protocol.T0, header + Lc + data + Le)
+    data, SW1, SW2 = transmit(hcard, Protocol.T0, header + Lc + data + Le)
     # Case 4S.1—Process aborted
     if SW1[0] == '6' and SW1[1] in '0456789ABCDEF':
-        return (data, SW1, SW2), F'Error—Process aborted ({SW1+SW2})'
+        raise Exception(
+            F'Error—Process aborted (data={data}, SW12={SW1}{SW2})')
     # Case 4S.2—Process completed
     if SW1+SW2 == '9000':
         # Send GET RESPONSE
@@ -372,18 +373,16 @@ def _send_apdu_T0_case_4s(hcard, header: str, Lc: str, data: str, Le: str):
         return _send_apdu_T0_case_2s(hcard, Protocol.T0, header[0:2] + 'C00000', _min(Ne, SW2))
     # Case 4S.4—SW12 = '62XY', or '63XY', or '9XYZ', except for '9000'
     if SW1.startswith('9') or SW1 in ['62', '63']:
-        return (data, SW1, SW2), None
+        return data, SW1, SW2
 
-    return (data, SW1, SW2), 'Unknown command case'
+    raise Exception(F'Unknown command case (data={data}, SW12={SW1}{SW2})')
 
 
 def get_status_change(hcontext, reader_states=None, timeout=None):
     """get_status_change()
     """
     if reader_states is None:
-        readers, err = list_readers(hcontext)
-        if err is not None:
-            return None, err
+        readers = list_readers(hcontext)
 
         reader_states = [(reader, _SCARD_STATE_UNAWARE)
                          for reader in readers]
@@ -396,12 +395,15 @@ def get_status_change(hcontext, reader_states=None, timeout=None):
     hresult, new_reader_states = _SCardGetStatusChange(
         hcontext, timeout, old_reader_states)
     if hresult != _SCARD_S_SUCCESS:
-        return None, F'Failed to get new reader states: _SCardGetErrorMessage(hresult)'
+        err = F'Failed to get new reader states ({_SCardGetErrorMessage(hresult)})'
+        logging.info(err)
+        raise Exception(err)
 
     new_reader_states = list(
         map(_convert_reader_state_from_scard, new_reader_states))
 
-    return new_reader_states, None
+    logging.info(F"new reader states: {', '.join(new_reader_states)}")
+    return new_reader_states
 
 
 def disconnect(hcard, disposition: Disposition):
@@ -411,12 +413,11 @@ def disconnect(hcard, disposition: Disposition):
 
     hresult = _SCardDisconnect(hcard, dw_disposition)
     if hresult != _SCARD_S_SUCCESS:
-        err = F'Failed to disconnect: _SCardGetErrorMessage(hresult)'
+        err = F'Failed to disconnect ({_SCardGetErrorMessage(hresult)})'
         logging.info(err)
-        return err
+        raise Exception(err)
 
     logging.info("Disconnected")
-    return None
 
 
 def release_context(hcontext):
@@ -426,10 +427,9 @@ def release_context(hcontext):
     if hresult != _SCARD_S_SUCCESS:
         err = F'Failed to release context: {_SCardGetErrorMessage(hresult)}'
         logging.info(err)
-        return err
+        raise Exception(err)
 
     logging.info("Context released")
-    return None
 
 #
 # Helper functions
